/*  main.c
 *
 *  Main body of xpsk31 application
 */

/*
 *  xpsk31: An application to transmit and receive
 *  PSK31 signals using a computer's sound card
 *
 *  Copyright (C) 2002  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "xpsk31.h"

/* xpsk31 main window */
GtkWidget *main_window;

/* Global widgets */
GtkWidget
  *gbl_popup_menu,
  *gbl_rx_scrolledwindow,
  *gbl_tx_scrolledwindow,
  *gbl_scope,
  *gbl_scope_label;

/* Pixbuffer for waterfall */
GdkPixbuf *gbl_wfall_pixbuf = NULL;
guchar *gbl_wfall_pixels;
gint
  gbl_wfall_rowstride,
  gbl_wfall_n_channels,
  gbl_wfall_width,
  gbl_wfall_height;

/* Text buffers for text viewers */
GtkTextBuffer
  *gbl_rx_text_buffer,
  *gbl_tx_text_buffer;

FILE
  /* File for recording QSO's */
  *qso_record_fp = NULL,
  /* File for ADIF QSO record */
  *log_adif_fp = NULL,
  /* File for staion QSO log  */
  *log_fp = NULL;

extern int
  dsp_in_fd,  /* File descriptor of DSP-in device  */
  dsp_out_fd, /* File descriptor of DSP-out device */
  mix_fd;     /* File descriptor of mixer device   */

int
  num_chan,		/* Stereo(=2) or Mono(=1) selection */
  chan_used,	/* Channel in use: left=0, right=1; */
  xmit_buf_size,/* Xmit DSP signal samples buffer   */
  recv_buf_size,/* Recv DSP signal samples buffer   */
  recv_buf_idx;	/* Index to Rx signal samples buffer*/

/* Signal samples buffers */
unsigned char *xmit_buffer;
unsigned char *recv_buffer;

char
  label[9][LABEL_LENGTH],	/* Labels for F1-F9 commands */
  *macro[9];				/* Macros attached to F1-F9  */

/* Runtime config data */
rc_data_t rc_data;

/* QSO data record */
extern qso_record_t qso_record;

/* Signal handler */
static void sig_handler( int signal );

/*------------------------------------------------------------------------*/

  int
main (int argc, char *argv[])
{
  /* Command line option returned by getopt() */
  int option;

  /* getopt() variables */
  extern char *optarg;
  extern int optind, opterr, optopt;

  /* New and old actions for sigaction() */
  struct sigaction sa_new, sa_old;


  /* Initialize new actions */
  sa_new.sa_handler = sig_handler;
  sigemptyset( &sa_new.sa_mask );
  sa_new.sa_flags = 0;

  /* Register function to handle signals */
  sigaction( SIGINT,  &sa_new, &sa_old );
  sigaction( SIGSEGV, &sa_new, 0 );
  sigaction( SIGFPE,  &sa_new, 0 );
  sigaction( SIGTERM, &sa_new, 0 );
  sigaction( SIGABRT, &sa_new, 0 );

  /* Initialize GTK+ */
  gtk_set_locale ();
  gtk_init (&argc, &argv);

  /* Default mode is BPSK */
  SetFlag( MODE_BPSK );

  /*** Process command line options ***/
  while( (option = getopt(argc, argv, "hv") ) != -1 )
	switch( option )
	{
	  case 'h': /* Print usage and exit */
		Usage();
		exit( 0 );

	  case 'v': /* Print version and exit */
		fprintf( stderr, "%s %s\n", PACKAGE, VERSION );
		exit( 0 );

	  default: /* Print usage and exit */
		Usage();
		exit( -1 );

	} /* End of switch( option ) */

  /* Create main window and menu */
  main_window     = create_main_window();
  gbl_popup_menu  = create_main_menu();

  /* Get Rx text buffer and scroller */
  gbl_rx_text_buffer = gtk_text_view_get_buffer
	( GTK_TEXT_VIEW(lookup_widget(main_window, "rx_textview")) );
  gbl_rx_scrolledwindow = lookup_widget( main_window, "rx_scrolledwindow" );

  /* Get Tx text buffer and scroller */
  gbl_tx_text_buffer = gtk_text_view_get_buffer
	( GTK_TEXT_VIEW(lookup_widget(main_window, "tx_textview")) );
  gbl_tx_scrolledwindow = lookup_widget( main_window, "tx_scrolledwindow" );

  /* Get scope widget */
  gbl_scope = lookup_widget( main_window, "scope" );
  gbl_scope_label = lookup_widget( main_window, "scope_label" );

  /* Set labels to Xmit and Rcve buttons */
  {
	GtkWidget
	  *button,
	  *xmit_status,
	  *rcve_status;

	button = lookup_widget( main_window, "xmit" );
	xmit_status = gtk_label_new ("TRANSMIT");
	gtk_widget_show (xmit_status);
	gtk_container_add (GTK_CONTAINER (button), xmit_status);
	gtk_label_set_justify (GTK_LABEL (xmit_status), GTK_JUSTIFY_LEFT);
	g_object_set_data_full (G_OBJECT (main_window), "xmit_status",
		gtk_widget_ref (xmit_status),
		(GDestroyNotify) gtk_widget_unref);

	button = lookup_widget( main_window, "rcve" );
	rcve_status = gtk_label_new ("RECEIVE");
	gtk_widget_show (rcve_status);
	gtk_container_add (GTK_CONTAINER (button), rcve_status);
	gtk_label_set_justify (GTK_LABEL (rcve_status), GTK_JUSTIFY_LEFT);
	g_object_set_data_full (G_OBJECT (main_window), "rcve_status",
		gtk_widget_ref (rcve_status),
		(GDestroyNotify) gtk_widget_unref);
  }

  /* Initialize fft */
  Ifft_Init();

  /* Show main window */
  gtk_widget_show (main_window);

  /* Make cosine wave table */
  if( Make_Cosine_Table() && Make_Sine_Table() )
  {
	/* Load runtime config file, abort on error */
	gtk_idle_add( Load_Config, NULL );
	SetFlag( ENABLE_AFC );
  }

  gtk_main ();

  return 0;
}

/*-------------------------------------------------------------------*/

/*  Load_Config()
 *
 *  Loads the xpsk31rc configuration file
 */

  gboolean
Load_Config( gpointer data )
{
  /* For setting macro button labels */
  GtkWidget *button;
  char bname[7];

  int
	lb_idx, /* Label buffer index */
	mc_idx, /* Macro buffer index */
	ln_idx, /* Line buffer index  */
	ln_len, /* Line buffer length */
	tg_cnt, /* Count '<' and '>'  */
	idx;

  char
	rc_fpath[64], /* File path to xpsk31rc */
	line[81];     /* Buffer for Load_Line  */

  /* Config file pointer */
  FILE *xpsk31rc;

  /* For setting squelch level */
  GtkWidget *hscale;
  GtkAdjustment *adjustment;


  /* Setup file path to xpsk31rc */
  snprintf( rc_fpath, 64, "%s/xpsk31/xpsk31rc", getenv("HOME") );

  /* Open xpsk31rc file */
  xpsk31rc = fopen( rc_fpath, "r" );
  if( xpsk31rc == NULL )
  {
	perror( rc_fpath );
	Error_Dialog(
		"Failed to open xpsk31rc file\n"
		"Quit xpsk31 and correct", TRUE );
	return( FALSE );
  }

  /*** Read runtime configuration data ***/
  /* Read mixer device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Mixer Device" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.mixer, line, 14 );
  rc_data.mixer[14] = '\0';

  /* Read input DSP device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Input DSP Device" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.dsp_in, line, 14 );
  rc_data.dsp_in[14] = '\0';

  /* Read output DSP device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Output DSP Device" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.dsp_out, line, 14 );
  rc_data.dsp_out[14] = '\0';

  /* Read recording source, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Recording Source" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.rec_src, line, 9 );
  rc_data.rec_src[9] = '\0';

  /* Read Recording level device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Recording Level Device" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.lev_dev, line, 9 );
  rc_data.lev_dev[9] = '\0';

  /* Read Recording level, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Recording Level" ) != SUCCESS )
	return( FALSE );
  rc_data.rec_lev = atoi( line );

  /* Read Volume level, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Volume Level" ) != SUCCESS )
	return( FALSE );
  rc_data.vol_lev = atoi( line );

  /* Read mode, abort if EOF */
  if( Load_Line(line, xpsk31rc, "STEREO|MONO Mode" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.mode, line, 6 );
  rc_data.mode[6] = '\0';

  /* Read channel used, abort if EOF */
  if( Load_Line(line, xpsk31rc, "LEFT|RIGHT Channel" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.channel, line, 5 );
  rc_data.channel[5] = '\0';

  /* Read squelch threshold, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Squelch Threshold" ) != SUCCESS )
	return( FALSE );
  rc_data.sqlch_thr = atoi( line );

  /* Enter squelch level to hscale */
  hscale = lookup_widget( main_window, "squelch_value" );
  adjustment = gtk_range_get_adjustment( GTK_RANGE(hscale) );
  gtk_adjustment_set_value( adjustment, (gdouble)rc_data.sqlch_thr );

  /* Read tone duration, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Tone Duration" ) != SUCCESS )
	return( FALSE );
  rc_data.tone_dur = atoi( line );

  /* Read number of reversals, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Number of Reversals" ) != SUCCESS )
	return( FALSE );
  rc_data.num_rev = atoi( line );

  /* Read CW encoder speed, abort if EOF */
  if( Load_Line(line, xpsk31rc, "CW Encoder Speed" ) != SUCCESS )
	return( FALSE );
  rc_data.wpm = atoi( line );

  /* Read word wrap column, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Word Wrap Column" ) != SUCCESS )
	return( FALSE );
  rc_data.wwrap = atoi( line );

  /*** Setup default main menu items ***/
  /* Read FT847 serial port device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "FT847 Serial Port" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.serial1, line, 14 );
  rc_data.serial1[14] = '\0';

  /* Read FT847 CAT enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, "FT847 CAT Enable" ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
	SetFlag( ENABLE_CAT_847 );
  else
	if( strcmp(line, "no") == 0 )
	  ClearFlag( ENABLE_CAT_847 );
	else
	{
	  Error_Dialog(
		  "Error reading xpsk31rc\n"
		  "Unrecognized menu option\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Read FT857 serial port device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "FT857 Serial Port" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.serial2, line, 14 );
  rc_data.serial2[14] = '\0';

  /* Read FT857 CAT enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, "FT857 CAT Enable" ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
	SetFlag( ENABLE_CAT_857 );
  else
	if( strcmp(line, "no") == 0 )
	  ClearFlag( ENABLE_CAT_857 );
	else
	{
	  Error_Dialog(
		  "Error reading xpsk31rc\n"
		  "Unrecognized menu option\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Try to auto-detect transceiver */
  if( Load_Line(line, xpsk31rc, "CAT Auto Detect" ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
	SetFlag( ENABLE_CAT );
  else
	if( strcmp(line, "no") == 0 )
	  ClearFlag( ENABLE_CAT );
	else
	{
	  Error_Dialog(
		  "Error reading xpsk31rc\n"
		  "Unrecognized menu option\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Read RTS-Control serial port device, abort if EOF */
  if( Load_Line(line, xpsk31rc, "RTS-Control Serial Port" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.serial3, line, 14 );
  rc_data.serial3[14] = '\0';

  /* Read RTS-Control enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, "RTS-Control Enable" ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
	SetFlag( ENABLE_RTS_CTL );
  else
	if( strcmp(line, "no") == 0 )
	  ClearFlag( ENABLE_RTS_CTL );
	else
	{
	  Error_Dialog(
		  "Error reading xpsk31rc\n"
		  "Unrecognized RTS-Control option\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Read Record QSO enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Record QSOs Enable" ) != SUCCESS )
	return( FALSE );

  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "record_qsos")), TRUE );
	SetFlag( RECORD_QSO );
  }
  else
	if( strcmp(line, "no") == 0 )
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "record_qsos")), FALSE );
	else
	{
	  Error_Dialog(
		  "Error reading xpsk31rc\n"
		  "Unrecognized menu option\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Read Capitalize enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Capitalize Enable" ) != SUCCESS )
	return( FALSE );

  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "capitalize_letters")), TRUE );
	SetFlag( CAPITAL_LETTERS );
  }
  else
	if( strcmp(line, "no") == 0 )
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "capitalize_letters")), FALSE );
	else
	{
	  Error_Dialog(
		  "Error reading xpsk31rc\n"
		  "Unrecognized menu option\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Read operator callsign, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Operator Callsign" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.call, line, 14 );
  rc_data.call[14] = '\0';

  /* Read operator name, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Operator Name" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.name, line, 12 );
  rc_data.name[12] = '\0';

  /* Read Zone, abort if EOF */
  if( Load_Line(line, xpsk31rc, "ITU Zone" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.zone, line, 14 );
  rc_data.zone[14] = '\0';

  /* Read QTH, abort if EOF */
  if( Load_Line(line, xpsk31rc, "QTH Name" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.qth, line, 12 );
  rc_data.qth[12] = '\0';

  /* Read QTH locator, abort if EOF */
  if( Load_Line(line, xpsk31rc, "QTH Locator" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.loc, line, 6 );
  rc_data.loc[6] = '\0';

  /* Read Transmitter name, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Transmitter Name" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.tx, line, 11 );
  rc_data.tx[11] = '\0';

  /* Read Transmitter power, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Transmitter Power" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.tx_power, line, 4 );
  rc_data.tx_power[4] = '\0';

  /* Read Transmitter antenna, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Transmitter Antenna" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.tx_ant, line, 15 );
  rc_data.tx_ant[15] = '\0';

  /* Read Receiver name, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Receiver Name" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.rx, line, 11 );
  rc_data.rx[11] = '\0';

  /* Read Receiver N.Fig, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Receiver Noise Figure" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.rx_nfig, line, 4 );
  rc_data.rx_nfig[4] = '\0';

  /* Read Receiver antenna, abort if EOF */
  if( Load_Line(line, xpsk31rc, "Receiver Antenna" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.rx_ant, line, 15 );
  rc_data.rx_ant[15] = '\0';

  /* Read propagation medium */
  if( Load_Line(line, xpsk31rc, "Propagation Medium" ) != SUCCESS )
	return( FALSE );
  strncpy( qso_record.via, line, 12 );
  qso_record.via[12] = '\0';

  /* Read default remarks on QSO */
  if( Load_Line(line, xpsk31rc, "Default Remarks" ) != SUCCESS )
	return( FALSE );
  strncpy( qso_record.remarks, line, 45 );
  qso_record.remarks[45] = '\0';

  /*** Read macro labels from rc file ***/
  /* Null buffer pointers just in case */
  for( idx = 0; idx < NUM_OF_LABELS; idx++ )
	macro[idx] = NULL;

  /* Preload a line */
  if( Load_Line(line, xpsk31rc, "Macro Labels" ) == SUCCESS )
	for( lb_idx = 0; lb_idx < NUM_OF_LABELS; lb_idx++ )
	{
	  /* Abort if first line is not a label */
	  if( line[0] != '[' )
	  {
		fprintf( stderr, "xpsk31: First Macro line not a label\n%s\n", line );
		Error_Dialog(
			"Error reading xpsk31rc\n"
			"First Macro line not a label\n"
			"Quit xpsk31 and correct", TRUE);
		return( FALSE );
	  }

	  /* Look for closing ']' */
	  ln_idx = 1;
	  while( (ln_idx <= LABEL_LENGTH) && (line[ln_idx] != ']') )
		ln_idx++;

	  /* Abort if label string > LABEL_LENGTH char + [] */
	  if( ln_idx > LABEL_LENGTH )
	  {
		fprintf( stderr, "xpsk31: Macro label too long\n%s\n", line );
		Error_Dialog(
			"Error reading xpsk31rc\n"
			"Macro label too long\n"
			"Quit xpsk31 and correct", TRUE);
		return( FALSE );
	  }

	  /* Enter label string if all OK */
	  snprintf( label[lb_idx], LABEL_LENGTH, "(F%d)", lb_idx+1 );
	  line[strlen(line)-1] = '\0';
	  strncat( label[lb_idx], &line[1], LABEL_LENGTH );

	  /* Set macro button labels */
	  snprintf( bname, 7, "macro%d", lb_idx+1 );
	  button = lookup_widget( main_window, bname );
	  gtk_button_set_label( GTK_BUTTON(button), label[lb_idx] );

	  /*** Read macros ***/
	  mc_idx = tg_cnt = 0;
	  do
	  {
		if( Load_Line(line, xpsk31rc, "Macros") != SUCCESS )
		  return( FALSE );;

		/* Read labels if line starts with '[' */
		if( line[0] == '[' ) break;

		/* Allocate memory to macro buffer */
		ln_len = strlen( line );

		/* Ignore | end-of-macro delimeter */
		if( line[ln_len-1] == '|' ) ln_len--;

		macro[lb_idx] = realloc(macro[lb_idx], ln_len + mc_idx + 2);

		/* Concatenate lines to macro */
		for( ln_idx = 0; ln_idx < ln_len; ln_idx++ )
		{
		  macro[lb_idx][mc_idx++] = line[ln_idx];

		  /* Count tag delimiters */
		  if( (line[ln_idx] == '<') ||
			  (line[ln_idx] == '>') )
			tg_cnt++;
		}

		/* Terminate Macro with LF and null */
		macro[lb_idx][mc_idx++] = LF;
		macro[lb_idx][mc_idx] = '\0';
	  } /* do */
	  while( line[ln_len] != '|'  );

	  /* Abort if tag delimiters are odd number */
	  if( tg_cnt & 1 )
	  {
		fprintf( stderr, "xpsk31: Malformed tag in Macro\n%s\n", line );
		Error_Dialog(
			"Error reading xpsk31rc\n"
			"Malformed tag in Macro\n"
			"Quit xpsk31 and correct", TRUE );
		return( FALSE );
	  }

	} /* for( lb_idx = 0; lb_idx < NUM_OF_LABELS; lb_idx++ ) */

  /*** Do the global config variables ***/
  /* Set stereo or mono mode */
  if( strcmp(rc_data.mode, "STEREO") == 0 )
	num_chan = 2;
  else
	if( strcmp(rc_data.mode, "MONO") == 0 )
	  num_chan = 1;
	else
	{
	  fprintf( stderr, "xpsk31: Sound card Mode invalid\n%s\n", rc_data.mode );
	  Error_Dialog(
		  "Error reading xpsk31rc file\n"
		  "Sound card Mode invalid\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Channel in use left or right */
  if( strcmp(rc_data.channel, "RIGHT") == 0 )
	chan_used = 1;
  else
	if( strcmp(rc_data.channel, "LEFT") == 0 )
	  chan_used = 0;
	else
	{
	  fprintf( stderr, "xpsk31: Channel in Use invalid\n%s\n", rc_data.mode );
	  Error_Dialog(
		  "Error reading xpsk31rc file\n"
		  "Channel in Use invalid\n"
		  "Quit xpsk31 and correct", TRUE );
	  return( FALSE );
	}

  /* Set up CAT if enabled */
  if( isFlagSet(ENABLE_CAT_847) )
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "ft847")), TRUE );
  else
	if( isFlagSet(ENABLE_CAT_857) )
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "ft857")), TRUE );
	else
	  if( isFlagSet(ENABLE_CAT) ) /* Auto-detect transceiver */
	  {
		ClearFlag( ENABLE_CAT );
		Close_Tcvr_Serial();
		ClearFlag( ENABLE_CAT_857 | SERIAL_IO_ERROR );
		SetFlag( ENABLE_CAT_847 );
		Open_Tcvr_Serial();

		if( isFlagSet(ENABLE_CAT) )
		  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
				lookup_widget(gbl_popup_menu, "ft847")), TRUE );
		else
		{
		  ClearFlag( ENABLE_CAT_847 | SERIAL_IO_ERROR );
		  SetFlag( ENABLE_CAT_857 );
		  Open_Tcvr_Serial();
		  if( isFlagSet(ENABLE_CAT) )
			gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
				  lookup_widget(gbl_popup_menu, "ft857")), TRUE );
		  else
		  {
			ClearFlag( ENABLE_CAT_857 | SERIAL_IO_ERROR );
			gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
				  lookup_widget(gbl_popup_menu, "none")), TRUE );
		  }
		} /* isFlagSet(ENABLE_CAT) */
	  } /* if( isFlagSet(ENABLE_CAT) */
		if( isFlagSet(ENABLE_RTS_CTL) )
		  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
				lookup_widget(gbl_popup_menu, "rts_ctl")), TRUE );

  /* Size of recv samples buffer depends on stereo/mono mode */
  recv_buf_size = (int)((double)DSP_SPEED/BAUD_RATE) * num_chan;
  /* Index to recv samples buffer (set to end) */
  recv_buf_idx  = recv_buf_size;

  /* Allocate memory to recv samples buffer */
  recv_buffer = malloc( recv_buf_size );
  if( recv_buffer == NULL )
  {
	Error_Dialog(
		"A memory allocation request failed\n"
		"Quit xpsk31 and correct", TRUE );
	return( FALSE );
  }

  /* Size of xmit samples buffer depends on stereo/mono mode */
  if( rc_data.tone_dur > 0 )
	xmit_buf_size =
	  num_chan * (DSP_SPEED * rc_data.tone_dur + rc_data.psk_elem2);
  else
	xmit_buf_size = 2 * rc_data.psk_elem2;

  /* Allocate memory to xmit samples buffer */
  xmit_buffer = malloc( xmit_buf_size );
  if( xmit_buffer == NULL )
  {
	Error_Dialog(
		"A memory allocation request failed\n"
		"Quit xpsk31 and correct", TRUE );
	return( FALSE );
  }

  /* Clean up xmit buffer */
  for( idx = 0; idx < xmit_buf_size; idx++ )
	xmit_buffer[idx] = 0x80;

  fclose( xpsk31rc );

  return( FALSE );

} /* End of Load_Config() */

/*------------------------------------------------------------------*/

/*  Load_Line()
 *
 *  Loads a line from a file, aborts on failure. Lines beginning
 *  with a '#' are ignored as comments. At the end of file EOF is
 *  returned. Lines assumed maximum 80 characters long.
 */

  int
Load_Line( char *buff, FILE *pfile, char *messg )
{
  int
	num_chr, /* Number of characters read, excluding lf/cr */
	chr;     /* Character read by getc() */
  char err_msg[128];

  /* Prepare error message */
  snprintf( err_msg, 128,
	  "Error reading %s\n"
	  "Premature EOF (End Of File)", messg );

  /* Clear buffer at start */
  buff[0] = '\0';
  num_chr = 0;

  /* Get next character, return error if chr = EOF */
  if( (chr = fgetc(pfile)) == EOF )
  {
	fprintf( stderr, "xpsk31: %s\n", err_msg );
	Error_Dialog( err_msg, TRUE );
	return( EOF );
  }

  /* Ignore commented lines, white spaces and eol/cr */
  while(
	  (chr == '#') ||
	  (chr == ' ') ||
	  (chr == HT ) ||
	  (chr == CR ) ||
	  (chr == LF ) )
  {
	/* Go to the end of line (look for LF or CR) */
	while( (chr != CR) && (chr != LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xpsk31: %s\n", err_msg );
		Error_Dialog( err_msg, TRUE );
		return( EOF );
	  }

	/* Dump any CR/LF remaining */
	while( (chr == CR) || (chr == LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xpsk31: %s\n", err_msg );
		Error_Dialog( err_msg, TRUE );
		return( EOF );
	  }

  } /* End of while( (chr == '#') || ... */

  /* Continue reading characters from file till
   * number of characters = 80 or EOF or CR/LF */
  while( num_chr < 80 )
  {
	/* If LF/CR reached before filling buffer, return line */
	if( (chr == LF) || (chr == CR) ) break;

	/* Enter new character to line buffer */
	buff[num_chr++] = chr;

	/* Get next character */
	if( (chr = fgetc(pfile)) == EOF )
	{
	  /* Terminate buffer as a string if chr = EOF */
	  buff[num_chr] = '\0';
	  return( SUCCESS );
	}

	/* Abort if end of line not reached at 80 char. */
	if( (num_chr == 80) && (chr != LF) && (chr != CR) )
	{
	  /* Terminate buffer as a string */
	  buff[num_chr] = '\0';
	  snprintf( err_msg, 128,
		  "Error reading %s\n"
		  "Line longer than 80 characters", messg );
	  fprintf( stderr, "xpsk31: %s\n%s\n", err_msg, buff );
	  Error_Dialog( err_msg, TRUE );
	  return( ERROR );
	}

  } /* End of while( num_chr < max_chr ) */

  /* Terminate buffer as a string */
  buff[num_chr] = '\0';

  return( SUCCESS );

} /* End of Load_Line() */

/*------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n",
	  "Usage: xpsk31 [-hv]" );

  fprintf( stderr, "%s\n",
	  "       -h: Print this usage information and exit." );

  fprintf( stderr, "%s\n",
	  "       -v: Print version number and exit." );

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleans up before quitting
 */

  void
Cleanup( void )
{
  /* Save outstanding records */
  if( isFlagSet(SAVE_RECORD) )
	Save_QSO_Record();

  /* Close open files */
  if( qso_record_fp != NULL )
  {
	fclose( qso_record_fp );
	qso_record_fp = NULL;
  }

  if( log_adif_fp != NULL )
  {
	fclose( log_adif_fp );
	log_adif_fp = NULL;
  }

  if( log_fp != NULL )
  {
	fclose( log_fp );
	log_fp = NULL;
  }

  /* Release sound card */
  if( dsp_in_fd > 0 )
  {
	close( dsp_in_fd );
	dsp_in_fd = 0;
  }

  if( dsp_out_fd > 0 )
  {
	close( dsp_out_fd );
	dsp_out_fd = 0;
  }

  if( mix_fd > 0 )
  {
	close( mix_fd );
	mix_fd = 0;
  }

  /* Release sound card and Tcvr */
  ClearFlag
	(
	 SOUND_CARD_SETUP |
	 TRANSMIT_MODE    |
	 TRANSMIT_MACRO   |
	 TRANSMIT_KEYBD   |
	 RECEIVE_MODE
	);

  Set_Tcvr_Mode();
  Close_Tcvr_Serial();

} /* Cleanup( void ) */

/*------------------------------------------------------------------------*/

/*  sig_handler()
 *
 *  Signal Action Handler function
 */

static void sig_handler( int signal )
{
  /* Cleanup and abort */
  Cleanup();

  fprintf( stderr, "\n" );
  switch( signal )
  {
	case SIGINT :
	  fprintf( stderr, "%s\n",  "xpsk31: Exiting via User Interrupt" );
	  exit( signal );

	case SIGSEGV :
	  fprintf( stderr, "%s\n",  "xpsk31: Segmentation Fault" );
	  exit( signal );

	case SIGFPE :
	  fprintf( stderr, "%s\n",  "xpsk31: Floating Point Exception" );
	  exit( signal );

	case SIGTERM :
	  fprintf( stderr, "%s\n",  "xpsk31: Termination Request received" );
	  exit( signal );

	case SIGABRT :
	  fprintf( stderr, "%s\n",  "xpsk31: Abort Signal received" );
	  exit( signal );

  } /* switch( signal ) */

} /* End of sig_handler() */

/*------------------------------------------------------------------------*/

/* Functions for testing and setting/clearing flow control flags
 *
 *  See xpsk31.h for definition of flow control flags
 */

/* An int variable holding the single-bit flags */
static int Flags = 0;

  int
isFlagSet( int flag )
{
  return( (Flags & flag) == flag );
}

  int
isFlagClear( int flag )
{
  return( (~Flags & flag) == flag );
}

  void
SetFlag( int flag )
{
  Flags |= flag;
}

  void
ClearFlag( int flag )
{
  Flags &= ~flag;
}

  void
ToggleFlag( int flag )
{
  Flags ^= flag;
}

/*------------------------------------------------------------------------*/
